/*
** Module   : HQLHelper.java
** Abstract : Helper object for random access queries
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------------Description-------------------------
** 001 ECF 20051007   @23251 Created initial version. Generates all HQL
**                           submitted on behalf of random access queries.
**                           Augments the base where clause of a query to
**                           provide the single-record retrieval and
**                           navigation semantic of Progress FIND and FOR
**                           queries.
** 002 ECF 20051111   @23379 Cleaned up exception handling. Removed throws
**                           clauses where possible, due to RecordBuffer
**                           changes.
** 003 ECF 20060114   @23916 Removed references to NameConverter. This is
**                           a conversion-only class. Use StringHelper
**                           instead.
** 004 ECF 20060204   @24357 Added temp table support. Account for temp
**                           table multiplexing in HQL statements.
** 005 ECF 20060306   @25021 Added support for foreign relation joins.
**                           Requires additional preprocessing of the
**                           where clause to inject a join phrase.
** 006 ECF 20060316   @25099 Integrated HQLPreprocessor. Rewrites certain
**                           base where clauses before augmenting them.
**                           Targets for preprocessing are where clauses
**                           which reference properties of a composite
**                           DMO inner class stored in an associated list.
** 007 ECF 20060529   @26592 Fixed generation of augmented where clauses
**                           for equality matches of text. These were not
**                           being properly uppercased.
** 008 ECF 20060602   @26941 Removed select clause for DMO ID. Persistence
**                           class has been optimized to retrieve records
**                           in a single step, rather than querying the ID
**                           first, then loading in a second step.
** 009 ECF 20060712   @28048 Migrated utility code to DBUtils class.
** 010 ECF 20060728   @28266 Rolled back #008 (@26941). Required for more
**                           reliable locking. The use of the second level
**                           cache should negate any loss of performance.
** 011 ECF 20060802   @28351 Moved inner class SortComponent to its own
**                           top-level class, SortCriterion. Related
**                           functionality in the HQLHelper constructor
**                           also migrated to the new class.
** 012 ECF 20060831   @29086 Simplified retrieval of schema name in
**                           factory method obtain().
** 013 ECF 20070228   @32244 Enable differentiation between ID-only
**                           projection queries and full queries. The
**                           former returns only the primary key ID of a
**                           record. The latter fetches the full DMO.
** 014 ECF 20070328   @32644 Changed to support HQLPreprocessor caching.
**                           Use factory method to retrieve an instance of
**                           the preprocessor rather than invoking the
**                           constructor directly.
** 015 ECF 20070502   @33369 Changed to support HQLPreprocessor changes.
**                           The signature of that class' primary factory
**                           has changed.
** 016 ECF 20070827   @35000 Optimized certain temp table queries.
**                           Eliminate multiplex ID check from where
**                           clause where possible and hard code multiplex
**                           ID when this test is necessary, rather than
**                           using a query substitution parameter. Augment
**                           order by clauses with leading multiplex ID.
** 017 ECF 20071206   @36268 Added sortIndex field and getSortIndex()
**                           method. Tracks SortIndex used by associated
**                           queries. Integrated generics.
** 018 ECF 20080310   @37475 Added args field. Needed by HQLPreprocesor to
**                           derive data type information for query
**                           substitution parameters.
** 019 ECF 20080403   @38149 Minor cleanup. Removed dmoIface from instance
**                           variables and method signatures.
** 020 SVL 20080421   @38081 The type of the "join" field was changed from
**                           DynamicJoin to AbstractJoin.
** 021 ECF 20080606   @38605 Added support for parallel dirty database HQL
**                           in HQLBundles. This is required when there
**                           are different dialects in use between the
**                           primary and dirty database.
** 022 ECF 20081009   @40077 Integrated new support for inlining query
**                           substitution parameters. Queries supported by
**                           this class don't actually have their
**                           parameters inlined, but changes were needed
**                           to integrate HQLPreprocessor API changes and
**                           the ParameterIndices class.
** 023 ECF 20081107  @40404  Moved multiplex ID position in order by
**                           clause. It used to be the first element, but
**                           that sometimes resulted in a less efficient
**                           index being selected for a query plan. Now it
**                           is either the last element, or second to last
**                           (if the primary key is part of the clause).
** 024 SVL 20090118  @41168  HQL helpers which have inlined preprocessors
**                           are not stored in cache.
** 025 SVL 20090122  @41197  Do not look into cache if we have a query
**                           with unknown parameter(s).
** 026 ECF 20090302  @41596  Fixed order by clause for temp tables. In
**                           cases where both multiplex ID and primary key
**                           ID were needed, the order of these two did
**                           not match their order in temp table indices.
**                           Also reduced memory use of HQLBundle cache.
** 027 GES 20090424   @41940 Import change.
** 028 ECF 20090602   @42583 Expose primary HQLPreprocessor.
** 029 ECF 20090623   @42942 Partially backed out #016 (@35000). Replaced hard
**                           coded multiplex IDs with query substitution
**                           parameter placeholders. Some use cases generate
**                           many similar queries which are differentiated
**                           only by the hard coded multiplex IDs. This was
**                           bloating the query cache in the Persistence
**                           Context class.
** 030 ECF 20131001          Added metadata support.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.*;
import org.apache.commons.logging.*;
import org.hibernate.type.Type;
import org.hibernate.dialect.Dialect;
import com.goldencode.util.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;

/**
 * Helper class which provides services to random access style queries.
 * These include composing HQL statements for all possible query navigation
 * and creating an array of Hibernate <code>Type</code> objects to match a
 * substitution parameter array.
 * <p>
 * This class is the work horse behind {@link RandomAccessQuery}, generating
 * one or more HQL select statements for each type of navigation/retrieval
 * request issued by that query.  An instance of {@link HQLBundle} is created
 * lazily for each such command type received from the query.  These instances
 * are cached to speed subsequent lookups within the current context.
 * <p>
 * Given a {@link RecordBuffer}, an optional where clause, an optional sort
 * clause, an optional {@link DynamicJoin}, and an indicator whether this
 * query is a projection query which fetches only the primary key of a
 * record, this class will:
 * <ul>
 *   <li>{@link HQLPreprocessor preprocess} the where clause to account for
 *       any implicit ANSI joins necessary to resolve certain complex uses of
 *       subscripts;
 *   <li>augment the base where clause if necessary to account for temporary
 *       table semantics;
 *   <li>augment the base where clause if necessary to embed any natural,
 *       foreign key joins;
 *   <li>analyze the sort clause in order to:
 *   <ul>
 *     <li>create an <code>order by</code> clause for the query;
 *     <li>augment the base where clause to allow ordered, single record
 *         navigation and retrieval;
 *   </ul>
 *   <li>lazily create an {@link HQLBundle} instance for each type of
 *       navigation/retrieval request processed by the query;
 *   <li>populate each <code>HQLBundle</code> instance with the full HQL
 *       statement(s), accessor method(s), and query substitution parameter
 *       indices needed to service a particular navigation/retrieval request.
 * </ul>
 */
final class HQLHelper
implements QueryConstants
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(HQLHelper.class);
   
   /** Cache of all active helper objects in the current context */
   private static final HQLHelperCache cache = new HQLHelperCache();
   
   /** Base where clause for associated query */
   private final String where;
   
   /** HQL where clause preprocessor */
   private final HQLPreprocessor hqlPreprocessor;
   
   /** HQL where clause preprocessor, using dirty database dialect */
   private final HQLPreprocessor altPreprocessor;
   
   /** DMO implementation class used by query */
   private final Class<?> dmoClass;
   
   /** Alias used to represent DMO instance in HQL statements */
   private final String dmoAlias;
   
   /** Whether query should return primary key ID only, or entire DMO */
   private final boolean idOnly;
   
   /** Multiplex ID of target buffer;  may be <code>null</code> */
   private final Integer multiplexID;
   
   /** Does sort clause need to be augmented with multiplex ID? */
   private final boolean multiplexSort;
   
   /** Helper object for dynamic join via a foreign key relation */
   private final AbstractJoin join;
   
   /** Components of a parsed order by clause */
   private final List<SortCriterion> sortCriteria;
   
   /** Components of a parsed order by clause, using dirty database dialect */
   private final List<SortCriterion> altCriteria;
   
   /** Cache of HQL bundles, indexed by retrieval request type constant */
   private final HQLBundle[] bundleCache = new HQLBundle[5];
   
   /** Sort index used by query */
   private final SortIndex sortIndex;
   
   /** Order by clause for query, adjusted for navigation direction */
   private String orderBy = null;
   
   /** Order by clause for dirty database use */
   private String altOrderBy = null;
   
   /**
    * Constructor which sets fundamental data upon which final HQL statements
    * will be based.  Parses the base order by clause <code>sort</code> into
    * its property and directional components, which are later used to augment
    * the base where clause for single-record retrievals.
    *
    * @param   buffer
    *          Record buffer against which the current query operates.
    * @param   schema
    *          Name of schema associated with query's buffer.
    * @param   dmoClass
    *          DMO implementation class used by query.
    * @param   dmoAlias
    *          Alias used to represent DMO instance in HQL statements.
    * @param   idOnly
    *          If <code>true</code>, the HQL statements generated by this
    *          object will begin with a select clause which specifies only the
    *          record's primary key identifier;  if <code>false</code>, no
    *          select clause will be prepended.
    * @param   join
    *          Helper object for a join via a foreign key relation.  May be
    *          <code>null</code>.
    * @param   where
    *          Base where clause for associated query.
    * @param   sort
    *          Base order by clause for associated query.
    * @param   args
    *          Query substitution parameters.
    * 
    * @throws  PersistenceException
    *          if there is any error parsing the given <code>sort</code>
    *          clause.
    */
   private HQLHelper(RecordBuffer buffer,
                     String schema,
                     Class<?> dmoClass,
                     String dmoAlias,
                     boolean idOnly,
                     AbstractJoin join,
                     String where,
                     String sort,
                     Object[] args)
   throws PersistenceException
   {
      this.dmoClass = dmoClass;
      this.dmoAlias = dmoAlias;
      this.idOnly = idOnly;
      this.join = join;
      this.where = where;
      this.hqlPreprocessor = HQLPreprocessor.get(where, buffer, args, false);
      this.sortCriteria = SortCriterion.parse(sort, buffer);
      this.sortIndex = SortIndex.get(sort, buffer);
      boolean hasWhere = (where != null && where.length() > 0);
      
      // Multiplex ID semantics are required in the where clause if we have
      // either a multiplexed temp table, or we have a where clause in a temp
      // table query.
      this.multiplexID = (buffer.isMultiplexed() || hasWhere
                          ? buffer.getMultiplexID()
                          : null);
      
      // The order by clause must be augmented with a multiplex ID for any
      // temp table query.
      this.multiplexSort = (buffer.getMultiplexID() != null);
      
      // If dirty database dialect differs from that of primary dialect, we
      // must generate an alternative SortCriteria list and HQLPreprocessor
      // (non-temp-tables only).
      Database primaryDB = buffer.getDatabase();
      Database dirtyDB = new Database(primaryDB, Database.Type.DIRTY);
      Dialect primaryDialect = DatabaseManager.getDialect(primaryDB);
      Dialect dirtyDialect = DatabaseManager.getDialect(dirtyDB);
      
      if (DatabaseManager.isTempTable(dmoClass) ||
          dirtyDialect == null ||
          dirtyDialect.equals(primaryDialect))
      {
         this.altCriteria = this.sortCriteria;
         this.altPreprocessor = this.hqlPreprocessor;
      }
      else
      {
         P2JDialect dialect = (P2JDialect) dirtyDialect;
         Class<?> dmoIface = buffer.getDMOInterface();
         this.altCriteria = SortCriterion.parse(dialect,
                                                sort,
                                                schema,
                                                dmoAlias,
                                                dmoClass,
                                                dmoIface,
                                                true);
         this.altPreprocessor = HQLPreprocessor.get(where,
                                                    dirtyDB,
                                                    dialect,
                                                    dmoClass,
                                                    args,
                                                    false);
      }
   }
   
   /**
    * Obtain an instance of an <code>HQLHelper</code> for the given buffer
    * and query criteria.  The instance is obtained from the cache if present,
    * otherwise, it is created and cached.
    *
    * @param   buffer
    *          Record buffer against which the current query operates.
    * @param   args
    *          Query substitution parameters.
    * @param   idOnly
    *          If <code>true</code>, the HQL statements generated by this
    *          object will begin with a select clause which specifies only the
    *          record's primary key identifier;  if <code>false</code>, no
    *          select clause will be prepended.
    * @param   where
    *          Base where clause supplied to the current query.
    * @param   sort
    *          Base order by clause supplied to the current query.
    * @param   join
    *          Helper object for a join via a foreign key relation.  May be
    *          <code>null</code>.
    *
    * @return  A new or cached instance of this class, specific to the given
    *          buffer and query criteria.
    */
   static HQLHelper obtain(RecordBuffer buffer,
                           Object[] args,
                           boolean idOnly,
                           String where,
                           String sort,
                           AbstractJoin join)
   {
      Class<?> joinIface = (join == null ? null : join.getInverseInterface());
      Type[] types = (args != null
                      ? DBUtils.makeTypeArray(args)
                      : new Type[] {});

      // Resolve all parameters and check whether we have any unknown
      // value(s).
      boolean unknowns = false;
      if (args != null)
      {
         for (Object parm : args)
         {
            if (parm instanceof FieldReference)
            {
                parm = ((FieldReference) parm).resolve();
            }

            if (parm instanceof BaseDataType &&
                ((BaseDataType) parm).isUnknown())
            {
               unknowns = true;
            }
         }
      }

      HQLHelper helper = null;

      if (!unknowns)
      {
         helper = cache.get(buffer, types, idOnly, where, sort, joinIface);
      }

      if (helper == null)
      {
         Class<?> dmoClass = buffer.getDMOImplementationClass();
         String schema = buffer.getSchema();
         String dmoAlias = buffer.getDMOAlias();
         where = StringHelper.safeTrim(where);
         
         try
         {
            helper = new HQLHelper(buffer,
                                   schema,
                                   dmoClass,
                                   dmoAlias,
                                   idOnly,
                                   join,
                                   where,
                                   sort,
                                   args);

            if (!helper.hqlPreprocessor.wasInlined())
            {
               cache.put(buffer,
                         types,
                         idOnly,
                         where,
                         sort,
                         joinIface,
                         helper);
            }
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
      }
      
      return helper;
   }
   
   /**
    * Return the sort index used by the helper's query.
    * 
    * @return  Sort index.
    */
   SortIndex getSortIndex()
   {
      return sortIndex;
   }
   
   /**
    * Return the primary HQL preprocessor used for the underlying query's
    * where clause.
    * 
    * @return  HQL preprocessor.
    */
   HQLPreprocessor getHQLPreprocessor()
   {
      return hqlPreprocessor;
   }
   
   /**
    * Create a bundle containing a single HQL statement and no getter methods,
    * designed to retrieve the first record which matches the base where
    * clause criteria.  Which record is first is determined by the sort order
    * specified in the base order by clause.
    * <p>
    * No augmented HQL statements are produced for the bundle.
    * <p>
    * The resulting bundle is cached within the current context;  subsequent
    * requests return the cached bundle.
    *
    * @return  HQL bundle containing the information necessary to retrieve
    *          the first record.
    */
   synchronized HQLBundle first()
   {
      HQLBundle bundle = bundleCache[FIRST];
      if (bundle == null)
      {
         bundle = new HQLBundle();
         
         // Use ascending sort.
         orderBy = buildSortClause(false, sortCriteria);
         
         // Use base where clause;  no augmentations except multiplex support
         // if temp table.
         String pWhere = preprocessWhere(null, hqlPreprocessor);
         bundle.addStatement(pWhere, false);
         
         // Final processing to complete HQL statement.
         postProcessBundle(bundle, true);
         
         if (hqlPreprocessor != altPreprocessor)
         {
            altOrderBy = buildSortClause(false, altCriteria);
            pWhere = preprocessWhere(null, altPreprocessor);
            bundle.addStatement(pWhere, true);
            postProcessBundle(bundle, false);
         }
         
         bundle.freeze();
         bundleCache[FIRST] = bundle;
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug("FIRST:  " + bundle);
         }
      }
      
      return bundle;
   }
   
   /**
    * Create a bundle containing a single HQL statement and no getter methods,
    * designed to retrieve the last record which matches the base where
    * clause criteria.  Which record is last is determined by the sort order
    * specified in the base order by clause.
    * <p>
    * No augmented HQL statements are produced for the bundle.
    * <p>
    * The resulting bundle is cached within the current context;  subsequent
    * requests return the cached bundle.
    *
    * @return  HQL bundle containing the information necessary to retrieve
    *          the last record.
    */
   synchronized HQLBundle last()
   {
      HQLBundle bundle = bundleCache[LAST];
      if (bundle == null)
      {
         bundle = new HQLBundle();
         
         // Use descending sort.
         orderBy = buildSortClause(true, sortCriteria);
         
         // Use base where clause;  no augmentations except multiplex support
         // if temp table.
         String pWhere = preprocessWhere(null, hqlPreprocessor);
         bundle.addStatement(pWhere, false);
         
         // Final processing to complete HQL statement.
         postProcessBundle(bundle, true);
         
         if (hqlPreprocessor != altPreprocessor)
         {
            altOrderBy = buildSortClause(true, altCriteria);
            pWhere = preprocessWhere(null, altPreprocessor);
            bundle.addStatement(pWhere, true);
            postProcessBundle(bundle, false);
         }
         
         bundle.freeze();
         bundleCache[LAST] = bundle;
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug("LAST:  " + bundle);
         }
      }
      
      return bundle;
   }
   
   /**
    * Create a bundle containing the HQL statement(s) and getter method(s)
    * necessary to retrieve the next record which matches the base where
    * clause criteria.  Which record is next is determined by the sort order
    * specified in the base order by clause, along with the sort-relevant
    * data in the current record of the associated buffer.
    * <p>
    * One or more augmented HQL statements are produced for the bundle.
    * There will be one getter method per statement.  The statements are
    * ordered from most to least specific.  Refer to the class description of
    * {@link HQLBundle} for further implementation details.
    * <p>
    * The resulting bundle is cached within the current context;  subsequent
    * requests return the cached bundle.
    *
    * @return  HQL bundle containing the information necessary to retrieve
    *          the next record.
    */
   synchronized HQLBundle next()
   {
      HQLBundle bundle = bundleCache[NEXT];
      if (bundle == null)
      {
         bundle = new HQLBundle();
         
         // Use ascending sort.
         orderBy = buildSortClause(false, sortCriteria);
         
         // Augment base where clause;  one variant per sort component.
         // Store DMO getter methods associated with substitution parameters.
         fillBundleLists(bundle, false, sortCriteria, true);
         
         // Final processing to complete HQL statement.
         postProcessBundle(bundle, true);
         
         // If necessary, augment HQL bundle with parallel statements that are
         // compatible with the dirty database's dialect.
         if (altCriteria != sortCriteria)
         {
            altOrderBy = buildSortClause(false, altCriteria);
            fillBundleLists(bundle, false, altCriteria, false);
            postProcessBundle(bundle, false);
         }
         
         bundle.freeze();
         bundleCache[NEXT] = bundle;
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug("NEXT:  " + bundle);
         }
      }
      
      return bundle;
   }
   
   /**
    * Create a bundle containing the HQL statement(s) and getter method(s)
    * necessary to retrieve the previous record which matches the base where
    * clause criteria.  Which record is previous is determined by the sort
    * order specified in the base order by clause, along with the
    * sort-relevant data in the current record of the associated buffer.
    * <p>
    * One or more augmented HQL statements are produced for the bundle.
    * There will be one getter method per statement.  The statements are
    * ordered from most to least specific.  Refer to the class description of
    * {@link HQLBundle} for further implementation details.
    * <p>
    * The resulting bundle is cached within the current context;  subsequent
    * requests return the cached bundle.
    *
    * @return  HQL bundle containing the information necessary to retrieve
    *          the previous record.
    */
   synchronized HQLBundle previous()
   {
      HQLBundle bundle = bundleCache[PREVIOUS];
      if (bundle == null)
      {
         bundle = new HQLBundle();
         
         // Use ascending sort.
         orderBy = buildSortClause(true, sortCriteria);
         
         // Augment base where clause;  one variant per sort component.
         // Store DMO getter methods associated with substitution parameters.
         fillBundleLists(bundle, true, sortCriteria, true);
         
         // Final processing to complete HQL statement.
         postProcessBundle(bundle, true);
         
         // If necessary, augment HQL bundle with parallel statements that are
         // compatible with the dirty database's dialect.
         if (altCriteria != sortCriteria)
         {
            altOrderBy = buildSortClause(true, altCriteria);
            fillBundleLists(bundle, true, altCriteria, false);
            postProcessBundle(bundle, false);
         }
         
         bundle.freeze();
         bundleCache[PREVIOUS] = bundle;
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug("PREVIOUS:  " + bundle);
         }
      }
      
      return bundle;
   }
   
   /**
    * Create a bundle containing a single HQL statement and no getter methods,
    * designed to retrieve a single, unique record which matches the base
    * where clause criteria.
    * <p>
    * No augmented HQL statements are produced for the bundle.  Since only a
    * single record is sought, sorting is irrelevant.
    * <p>
    * The resulting bundle is cached within the current context;  subsequent
    * requests return the cached bundle.
    *
    * @return  HQL bundle containing the information necessary to retrieve
    *          a unique record.
    */
   synchronized HQLBundle unique()
   {
      HQLBundle bundle = bundleCache[UNIQUE];
      if (bundle == null)
      {
         bundle = new HQLBundle();
         
         // Null out order by clause.
         orderBy = null;
         
         // Use base where clause;  no augmentations.
         String pWhere = preprocessWhere(null, hqlPreprocessor);
         bundle.addStatement(pWhere, false);
         postProcessBundle(bundle, true);
         
         if (hqlPreprocessor != altPreprocessor)
         {
            altOrderBy = null;
            pWhere = preprocessWhere(null, altPreprocessor);
            bundle.addStatement(pWhere, true);
            postProcessBundle(bundle, false);
         }
         
         bundle.freeze();
         bundleCache[UNIQUE] = bundle;
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug("UNIQUE:  " + bundle);
         }
      }
      
      return bundle;
   }
   
   /**
    * Preprocess the where clause to account for temp table multiplexing and
    * foreign relation joins.  This is performed on the intermediate form of
    * the base where clause as created by the {@link HQLPreprocessor}, but
    * before it is augmented to account for the sort criteria in {@link
    * #fillBundleLists}.
    * <p>
    * If multiplexing is called for, this will prepend a test for the proper
    * multiplex ID before the where clause, if any.
    * <p>
    * If a foreign relation join is called for, the proper HQL will be
    * generated to compare a local entity with a foreign one (the foreign
    * entity is represented by a query substitution character:
    * <code>?</code>).
    *
    * @param   buf
    *          Buffer into which to write HQL.  May be <code>null</code>.
    * @param   preproc
    *          HQL preprocessor to use for the base where clause.
    *
    * @return  The preprocessed where clause, or <code>null</code> if the
    *          <code>buf</code> parameter was not <code>null</code>.
    */
   private String preprocessWhere(StringBuilder buf, HQLPreprocessor preproc)
   {
      boolean multiplex = (multiplexID != null);
      boolean hasWhere = (where != null && where.length() > 0);
      boolean hasJoin = (join != null);
      
      if (!multiplex && !hasWhere && !hasJoin)
      {
         // No preprocessing required.
         return null;
      }
      
      boolean needReturn = (buf == null);
      
      if (needReturn)
      {
         buf = new StringBuilder("where ");
      }
      
      // Handle temp table multiplexing.
      if (multiplex)
      {
         if (hasWhere || hasJoin)
         {
            buf.append("(");
         }
         
         buf.append(dmoAlias);
         buf.append(".");
         buf.append(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
         buf.append(" = ?");

         if (hasWhere || hasJoin)
         {
            buf.append(") and ");
         }
      }
      
      // Handle a join.
      if (hasJoin)
      {
         if (multiplex || hasWhere)
         {
            buf.append("(");
         }
         
         buf.append(join.getHQL());
         
         if (multiplex || hasWhere)
         {
            buf.append(")");
         }
         if (hasWhere)
         {
            buf.append(" and ");
         }
      }
      
      // Embed the base where clause, if any.
      if (hasWhere)
      {
         if (multiplex || hasJoin)
         {
            buf.append("(");
         }
         buf.append(preproc.getHQL());
         if (multiplex || hasJoin)
         {
            buf.append(")");
         }
      }
      
      return (needReturn ? buf.toString() : null);
   }
   
   /**
    * Post-process the augmented where clauses stored during the first pass
    * of HQL bundle creation.  Post-processing involves prepending each where
    * clause with select and from clauses, and appending with an optional
    * order by clause.  The original where clauses in the bundle are replaced
    * with the resulting, full HQL statements.
    *
    * @param   bundle
    *          HQL bundle to be post-processed.
    * @param   primary
    *          <code>true</code> to process HQL statement for primary database
    *          dialect;  <code>false</code> to process for dirty database
    *          dialect.
    */
   private void postProcessBundle(HQLBundle bundle, boolean primary)
   {
      int size = bundle.statementsSize();
      for (int i = 0; i < size; i++)
      {
         StringBuilder buf = new StringBuilder();
         
         assembleSelectClause(buf);
         assembleFromClause(buf);
         
         String where = bundle.getStatement(i, !primary);
         if (where != null)
         {
            buf.append(where);
            buf.append(" ");
         }
         
         if (primary)
         {
            assembleOrderByClause(buf, orderBy);
            bundle.setStatement(i, buf.toString().trim(), false);
         }
         else
         {
            assembleOrderByClause(buf, altOrderBy);
            bundle.setStatement(i, buf.toString().trim(), true);
         }
      }
      
      // Store query substitution parameter indices.
      // TODO:  currently, these are shared, and not dialect specific, but
      // only because we do not inline parameters from this class.  Were this
      // to change, we would need to maintain separate ParameterIndices
      // objects for the primary and dirty database statements.
      if (primary)
      {
         bundle.setParameterIndices(hqlPreprocessor.getParameterIndices());
      }
   }
   
   /**
    * Compose an HQL compliant order by clause from the sort components
    * parsed from the base order by clause submitted at construction,
    * optionally inverting the sort direction of each component.  The list of
    * original components will have been augmented with the DMO's primary key,
    * if this is necessary to disambiguate among non-unique records.
    *
    * @param   invert
    *          <code>true</code> to invert sort direction (i.e., asc-->desc,
    *          desc-->asc);  else <code>false</code>.
    * @param   sortCrit
    *          List of <code>SortCriterion</code> objects appropriate to the
    *          database dialect.
    *
    * @return  An HQL compliant order by clause.
    */
   private String buildSortClause(boolean invert,
                                  List<SortCriterion> sortCrit)
   {
      StringBuilder buf = new StringBuilder("order by ");
      
      boolean mpidInjected = false;
      Iterator<SortCriterion> iter = sortCrit.iterator();
      int i = 0;
      for (; iter.hasNext(); i++)
      {
         if (i > 0)
         {
            buf.append(", ");
         }
         SortCriterion next = iter.next();
         
         // Inject multiplex ID for temp table before primary key, so as to
         // match the way these components are ordered in the table's index.
         if (!mpidInjected &&
             multiplexSort &&
             DatabaseManager.PRIMARY_KEY.equals(next.getUnqualifiedName()))
         {
            injectMultiplexSort(invert, buf);
            buf.append(", ");
            mpidInjected = true;
         }
         
         buf.append(next.toSortExpression(invert));
      }
      
      // Inject multiplex ID for temp table if not previously injected.
      if (!mpidInjected && multiplexSort)
      {
         if (i > 0)
         {
            buf.append(", ");
         }
         injectMultiplexSort(invert, buf);
      }
      
      return buf.toString();
   }
   
   /**
    * Inject a sort expression into the given buffer using the well-known
    * multiplex ID field name with the appropriate DMO alias, and a sort
    * direction.
    * 
    * @param   invert
    *          If <code>true</code> the sort direction is descending;  if
    *          <code>false</code>, it is ascending.
    * @param   buf
    *          String builder into which the expression is injected.
    */
   private void injectMultiplexSort(boolean invert, StringBuilder buf)
   {
      buf.append(dmoAlias);
      buf.append(".");
      buf.append(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
      buf.append(" ");
      buf.append(invert ? "desc" : "asc");
   }
   
   /**
    * Fill an HQL bundle's internal HQL statement and getter method lists
    * with the appropriate statements and methods, respectively.  The number
    * of statements and methods correlates directly with the number of sort
    * components managed by the associated query.  The HQL statements are
    * generated and stored in order from most specific to least specific,
    * such that as they are executed, the criteria become decreasingly
    * restrictive.
    * <p>
    * This method represents the first pass in a two step process to create
    * a set of full, HQL statements ({@link #postProcessBundle} represents
    * the second pass).  In this pass, only augmented where clauses (and
    * their associated getter methods) are stored in the bundle.  The base
    * where clause is augmented with a binary match phrase for each sort
    * component, each of which is AND'd together.  The last test in the
    * augmented where clause is always a range match, whereas the previous
    * tests are always equality matches.  The range match operator is
    * determined by the sort direction of the sort component, and the value
    * of the <code>invert</code> parameter.
    * <p>
    * For example, given the following query criteria, with
    * <code>invert</code> set to <code>false</code>...
    * <pre>
    *    DMO:        Customer
    *    alias:      customer
    *    where:      customer.number = ? and customer.creditLimit = ?
    *    order by:   customer.creditLimit asc, customer.name asc
    * </pre>
    * <p>
    * ...this method would produce the following list of augmented where
    * clauses...
    * <pre>
    * 1) where (customer.number = ? and customer.creditLimit = ?)
    *      and (customer.creditLimit = ? and customer.name = ? and customer.id > ?)
    *
    * 2) where (customer.number = ? and customer.creditLimit = ?)
    *      and (customer.creditLimit = ? and customer.name > ?)
    *
    * 3) where (customer.number = ? and customer.creditLimit = ?)
    *      and (customer.creditLimit > ?)
    * </pre>
    * <p>
    * ...and the following list of DMO "getter" methods...
    * <pre>
    * 1) Customer.getCreditLimit()
    * 2) Customer.getName()
    * 3) Customer.getId()
    * </pre>
    * <p>
    * Note that the above example assumes neither <code>creditLimit</code>
    * nor <code>name</code> (nor the combination of the two) represent a
    * unique constraint.  If this assumption was not valid, the method would
    * produce the following augmented where clauses instead...
    * <pre>
    * 1) where (customer.number = ? and customer.creditLimit = ?)
    *      and (customer.creditLimit = ? and customer.name > ?)
    *
    * 2) where (customer.number = ? and customer.creditLimit = ?)
    *      and (customer.creditLimit > ?)
    * </pre>
    * <p>
    * ...and the following list of DMO "getter" methods...
    * <pre>
    * 1) Customer.getCreditLimit()
    * 2) Customer.getName()
    * </pre>
    * <p>
    * If a temp table is involved in the where clause, and/or if a foreign
    * relation join between tables is required, additional preprocessing of
    * the where clause must occur.
    *
    * @param   bundle
    *          HQL bundle to be populated.
    * @param   invert
    *          <code>true</code> to invert sort direction (i.e., asc-->desc,
    *          desc-->asc) of each sort component when generating the range
    *          match portion of the augmented where clause;  else
    *          <code>false</code>.  Inverting the sort direction in this
    *          context means using the opposite range operator in the match
    *          test (<code>&gt;</code> becomes <code>&lt;</code> and vice
    *          versa).
    * @param   sortCrit
    *          List of <code>SortCriterion</code> objects appropriate to the
    *          database dialect.
    * @param   primary
    *          <code>true</code> to add generated HQL statement to the HQL
    *          bundle for use with the primary database;  <code>false</code>
    *          to add it for use with the dirty database.
    *
    * @see     #preprocessWhere
    */
   private void fillBundleLists(HQLBundle bundle,
                                boolean invert,
                                List<SortCriterion> sortCrit,
                                boolean primary)
   {
      String dmoName = dmoClass.getName();
      dmoName = dmoName.substring(dmoName.lastIndexOf(".") + 1);
      
      // Store methods.  These are shared between primary and dirty database
      // statements and are not dialect specific.  They must only be added
      // once.
      if (primary)
      {
         Iterator<SortCriterion> iter = sortCrit.iterator();
         while (iter.hasNext())
         {
            SortCriterion next = iter.next();
            Method getter = next.getMethod();
            if (getter != null)
            {
               bundle.addGetter(getter);
            }
         }
      }
      
      // Compose where clause variants.
      int size = sortCrit.size();
      StringBuilder prebuf = new StringBuilder();
      preprocessWhere(prebuf, (primary ? hqlPreprocessor : altPreprocessor));
      String prep = prebuf.toString();
      if (prep.trim().length() == 0)
      {
         prep = null;
      }
      
      // Create one augmented where clause variant for each sort component.
      for (int i = size - 1; i >= 0; i--)
      {
         StringBuilder buf = new StringBuilder("where ");
         
         if (prep != null)
         {
            buf.append("(");
            buf.append(prep);
            buf.append(") and (");
         }
         
         for (int j = 0; j <= i; j++)
         {
            SortCriterion next = sortCrit.get(j);
            if (next.getMethod() == null)
            {
               j--;
               continue;
            }
            
            if (j > 0)
            {
               buf.append(" and ");
            }
            
            buf.append(next.toWhereExpression((j == i), invert));
         }
         
         if (prep != null)
         {
            buf.append(")");
         }
         
         bundle.addStatement(buf.toString(), !primary);
      }
   }
   
   /**
    * Assemble the <code>select</code> clause for an HQL query.  Uses the
    * form:
    * <pre>
    *   select {dmo}.id
    * </pre>
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    */
   private void assembleSelectClause(StringBuilder buf)
   {
      if (idOnly)
      {
         buf.append("select ");
         buf.append(dmoAlias);
         buf.append(".id ");
      }
   }
   
   /**
    * Assemble the <code>from</code> clause for an HQL query.  Uses the form:
    * <pre>
    *   from {DmoImpl} as {dmo}
    *   [ join {dmo}.{property} as {dmo}_{property} [...] ]
    * </pre>
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    */
   private void assembleFromClause(StringBuilder buf)
   {
      String dmoName = dmoClass.getName();
      dmoName = dmoName.substring(dmoName.lastIndexOf(".") + 1);
      
      buf.append("from ");
      buf.append(dmoName);
      buf.append(" as ");
      buf.append(dmoAlias);
      buf.append(" ");
      
      // Add any required ANSI joins created by the HQL preprocessor.  These
      // are shared for primary and dirty database use, and are not dialect
      // specific.
      Iterator<String> iter = hqlPreprocessor.ansiJoins();
      while (iter.hasNext())
      {
         buf.append(iter.next());
         buf.append(" ");
      }
   }
   
   /**
    * Assemble the <code>order by</code> clause for an HQL query.  Uses the
    * form:
    * <pre>
    *   order by {dmo1} {asc/desc} [, {dmo2} {asc/desc} [...]]
    * </pre>
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    * @param   orderPhrase
    *          Order by phrase to appended to the buffer if not
    *          <code>null</code>.
    */
   private void assembleOrderByClause(StringBuilder buf, String orderPhrase)
   {
      if (orderPhrase != null)
      {
         buf.append(orderPhrase);
      }
   }
}
