/*
** Module   : RelationNode.java
** Abstract : Tree node for experimental analysis of natural relations
**            between database tables
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050325   @20998 Created initial version.
** 002 GES 20090518   @42400 Import change.
** 003 ECF 20131013          Implemented generics.
*/

package com.goldencode.p2j.schema;

import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.util.*;

/**
 * A node in a tree of table relation dependencies. This class was written
 * for experimental analysis of primary-foreign key relation candidates
 * among Progress tables. It probably is of little value for production use
 * and should be deprecated.
 */
class RelationNode
extends TreeNode
implements SchemaParserTokenTypes
{
   /** Table wrapped by this node */
   private Aast table = null;
   
   /** Names of index fields which comprise the relation */
   private Collection<String> key = null;
   
   /**
    * Constructor which accepts a table and collection of fields which
    * define this node's relation with another node's table.
    *
    * @param   table
    *          Table AST wrapped by this node.
    * @param   key
    *          Collection of field names which define the key between tables.
    */
   RelationNode(Aast table, Collection<String> key)
   {
      this.table = table;
      this.key = key;
   }
   
   /**
    * Set the relation key for this node.
    *
    * @param   key
    *          Collection of field names which define the key between tables.
    */
   void setKey(Collection<String> key)
   {
      this.key = key;
   }
   
   /**
    * Create a string representation of this node for debug output.
    *
    * @return  String representation of this node.
    */
   public String toString()
   {
      StringBuilder buf = new StringBuilder();
      
      int depth = getDepth();
      for (int i = 0; i < depth; i++)
      {
         buf.append("   ");
      }
      
      buf.append(table.getText());
      
      if (key == null)
      {
         return buf.toString();
      }
      
      buf.append(" [");
      
      Iterator<String> iter = key.iterator();
      for (int i = 0; iter.hasNext(); i++)
      {
         if (i > 0)
         {
            buf.append(", ");
         }
         
         String name = iter.next();
         
         Aast field = null;
         Aast next = (Aast) table.getFirstChild();
         while (field == null && next != null)
         {
            int type = next.getType();
            if (((type > BEGIN_FIELDTYPES && type < END_FIELDTYPES) ||
                 (type > BEGIN_METATYPES  && type < END_METATYPES)) &&
                 name.equalsIgnoreCase(next.getText()))
            {
               field = next;
            }
            
            next = (Aast) next.getNextSibling();
         }
         
         buf.append(name).append(" (").append(field.getSymbolicTokenType()).append(")");
      }
      
      buf.append(']');
      
      return buf.toString();
   }
}
